;;; getopt.scm --- POSIX getopt(2) option parsing
;;;
;;; Copyright (c) 2021 nytpu <alex [at] nytpu.com>
;;; SPDX-License-Identifier: MPL-2.0
;;; Home Page: <https://git.sr.ht/~nytpu/getopt-scheme>

;; Replace a key in an alist if it exists, otherwise add it to the alist.
(define (alist-replace key datum alist)
  (if (assoc key alist)
      (alist-cons key datum (alist-delete key alist))
      (alist-cons key datum alist)))

;; Convert a POSIX optstring into an alist where the car is the option
;; character and the \scheme{cdr} is \scheme{#t} (takes an argument) or
;; \scheme{#f} (doesn't take an argument).  Leading ':'s are ignored
;; since this implementation of getopt doesn't print anything for you and
;; doesn't return ':' and '?' like in the C function.
(define (optstring->alist optstring)
  (let internal ((ostr (if (and (>= (string-length optstring) 1)
                                (char=? (string-ref optstring 0) #\:))
                           (cdr (string->list optstring))
                           (string->list optstring)))
                 (oset '()))
    (if (null? ostr)
        oset
        (if (and (not (null? (cdr ostr)))
                 (char=? (cadr ostr) #\:))
            (internal (cddr ostr)
                      (alist-replace (car ostr) #t oset))
            (internal (cdr ostr)
                      (alist-replace (car ostr) #f oset))))))

(define (pseudo-command-line fake)
  (if (null? fake)
      (command-line)
      fake))

;; A note to people actually looking at this abomination, this is:
;; 1) the first time I've ever written anything non-trivial in Scheme, and;
;; 2) basically a transliteration of my Ada implementation of the algorithm;
;; so if it's very shitty and unidiomatic Scheme that's why.  Patches welcome!
;; (my Ada version: <https://git.sr.ht/~nytpu/getopt-ada>)

;;> \scheme{getopt} will parse the command line parameters of a Scheme program
;;> following the POSIX Utility Syntax Guidelines.
;;>
;;> Usage: \scheme{(getopt optstring virtual-command-line ...)}
;;>
;;> Where \var{optstring} is an options string as defined in POSIX.
;;> See
;;> \hyperlink[https://pubs.opengroup.org/onlinepubs/9699919799/functions/getopt.html]{the specification}
;;> for more details.  Note that leading ':'s are ignored since this
;;> implementation of getopt doesn't print anything for you and doesn't return
;;> ':' and '?' like in the C function; the appropriate error is always raised.
;;>
;;> Sample optstrings:
;;> \itemlist[
;;> \item{\code{"abcd"} - Accept flags a, b, c, and d; none of which take parameters.}
;;> \item{\code{"a:b:c:d:"} - Accept flags a, b, c, and d; all of which require parameters.}
;;> \item{\code{"W:xy:Z"} - accept flags W, x, y, and Z.  x and Z don't take parameters, W and y require parameters.}
;;> \item{\code{":ab:cd"} - Accept flags a, b, c, and d; b requires a parameter.}
;;> ]
;;>
;;> \var{virtual-command-line ...} is a series of strings as would be returned
;;> by \scheme{(command-line)}.  If there are no virtual command line arguments
;;> provided then the list returned by \scheme{(command-line)} is used.
;;>
;;> \scheme{getopt} will return an association list with the keys corresponding
;;> to each of the flags present.  If the flag does not accept a value then the
;;> value of the association will be \scheme{#t}, otherwise it will be a string
;;> containing the argument.
;;>
;;> The special key \scheme{'command} contains the name of the command.
;;>
;;> The special key \scheme{'@} contains the remaining non-option arguments, in
;;> the order they were present in on the command line.

(define (getopt ostr . cmdln)
  (let internal ((opts (list (cons 'command (car (pseudo-command-line cmdln)))))
                 (cmdln (cdr (pseudo-command-line cmdln)))
                 (optpos 1)
                 (in-options? #t)
                 (oset (optstring->alist ostr)))
    (if (null? cmdln)
        opts
        (if in-options?
            (let* ((optind (car cmdln))
                   (optind-len (string-length optind)))
              (cond
                ((or (< optind-len 1)
                     (not (char=? (string-ref optind 0) #\-)))
                 (internal opts cmdln optpos #f oset))
                ((string=? optind "--") (internal opts (cdr cmdln) 0 #f oset))
                ((>= optpos optind-len)
                 (internal opts (cdr cmdln) 1 in-options? oset))
                (else
                  (let* ((option (string-ref optind optpos))
                         (from-oset (assoc option oset)))
                    (if (not from-oset)
                        (error "unrecognized option" option)
                        (if (cdr from-oset)
                            (cond
                              ((< (+ optpos 1) optind-len)
                               (internal (alist-replace option
                                                        (string-copy optind (+ optpos 1))
                                                        opts)
                                         (cdr cmdln) 1 #t oset))
                              ((not (null? (cdr cmdln)))
                               (internal (alist-replace option (cadr cmdln)
                                                        opts)
                                         (cddr cmdln) 1 #t oset))
                              (else (error "option requires an argument"
                                           option)))
                            (internal (alist-replace option #t opts)
                                      cmdln (+ optpos 1) #t oset)))))))
            (alist-cons '@ cmdln opts)))))
